#!/tools/gdaac/COTS/perl-5.8.5/bin/perl

=head1 NAME

s4pm_run_algorithm - run a science algorithm within S4PM

=head1 SYNOPSIS

s4pm_run_algorithm.pl 
[B<-c>]
[B<-f> I<config_file>] 
[B<-q> I<config_file>]
[B<-s>]
[B<-l> I<logdir>]
[B<-m> I<msgdir>] 
[B<-e> I<expiration_time>] 
[B<-p> I<pgedir>] 
[B<-t> I<toolkit_home>] 
[B<-v> I<PGEversion>] 
[B<-F> I<FAILPGE_met_template>] 
[B<-o> I<originating_system>]
[B<-a> I<node_name>]
[B<-O>]
[B<-A>]
[B<-R>]
executable process_control_file

=head1 DESCRIPTION

This runs a Product Generation Executable within S4P.
The penultimate argument is the full path to the executable, and the last 
argument is the full path the PCF (process control file).
In S4P, the PCF is the input work order.

On completion, s4pm_run_algorithm.pl also gathers up the production history 
information, and outputs a number of work orders:

=over 4

=item EXPORT

Product Delivery Record file with output data.

=item UPDATE

UPDATE work order for Track Data, decrementing uses for input data.

=item REGISTER

Product Delivery Record file with output data for Register Data. This is output
unless the B<-q> option is specified to send data to the qc station.

=back

=head1 OPTIONS

=over 4

=item B<-A>

A temporary flag, indicates that s4pm_run_algorithm.pl should use Aqua 
behaviour for decrementing uses, specifically -1 for all input granules.
If not set, it uses Terra behaviour, where input L0 is decremented according
to the number of MOD01 granules created.

When this has been thoroughly tested, the -A option will be removed and all
input granules will be decremented by 1.

=item B<-a> I<node_name>

By default, the node name added to the output PDRs going to the Export station
are taken from the machine name using S4P::PDR::gethost(). To explictly set
the node name to something else, include this option.

=item B<-c>

Add checksums to the EXPORT work order.

=item B<-f> I<config_file>

Configuration file.  This is really necessary only if you have a browse
map that needs to be represented, in which case the map should be in
%cfg_data_browse_map, with data lun as key and browse lun
as value.  There can be only one browse file for a given data file

In for a penny, in for a pound, so all of the other arguments listed
below can also be specified in the configuration file.  The variable name is
simply $cfg_argument, where argument is the argument listed below.
However, anything specified on the command line will override anything in
the configuration file.

=item B<-q> I<config_file>

Apply Quality Control scripts found in config_file.

=item B<-e> I<expiration_time>

Expiration time for output PDR work orders.  This should be specified as
<n><units>, e.g. 2days.  Valid units are seconds, minutes, hours, 
days, weeks, months and years.

The default is 3days.

=item B<-o> I<originating_system>

Originating system to use in output PDR work orders.
The default is 'S4PM'.

=item B<-s>

Symlinks:  creates local symbolic links to input and output directories
to aid in debugging.

=item B<-m> I<msgdir>

Message directory for PGE.  The default is '../MSGS'.

=item B<-p> I<pge_dir>

Directory for algorithm executables.  Default is '../'.
This is needed for nested executables that expect to be in the same directory.

=item B<-t> I<toolkit_dir>

Directory where Toolkit lives, used for directories starting with '~'.

=item B<-l> I<logdir>

Directory to save runlogs (output by rusage program).

=item B<-v> I<pge_version>

Version of the algorithm.  If this is not specified, it will attempt to obtain 
a version from the B<-p> argument.  This is used only for metadata for the 
FAILPGE, you can get by without it so long as the algorithm doesn't fail 8-).

=item B<-F> I<failpge_met_template>

Metadata template file for FAILPGE data type.
Default is ../FAILPGE_met_template.cfg.

=item B<-o> I<originating_system>

Specify alternate originating system for output PDR work orders.

=item B<-O>

Run algorithm in on-demand setting.
This modifies the name of the output EXPORT work order to not include
the algorithm.

=item B<-R>

Re-read PCF after execution. 
This is useful if the PGE tweaks the PCF to put in the "real"
output filenames.

=back

=head1 SEE ALSO

QC(3)

=head1 AUTHOR

Chris Lynnes, NASA/GSFC, Code 610.2, Greenbelt, MD  20771.

=cut

################################################################################
# s4pm_run_algorithm.pl,v 1.5 2006/10/16 00:14:59 lynnes Exp
# -@@@ S4PM, Version Release-5_16_0
################################################################################

use strict;
use S4P;
use S4P::PDR;
use S4P::FileGroup;
use S4P::PCF;
use S4P::TimeTools;
use S4P::MetFile;
use S4PM;
use S4PM::QC;
use POSIX;
use File::Basename;
use File::Copy;
use Getopt::Std;
use Safe;
require 5.6.0;
use vars qw($opt_a $opt_c $opt_e $opt_f $opt_l $opt_m $opt_o $opt_O 
            $opt_p $opt_P $opt_q $opt_R $opt_s $opt_t $opt_v $opt_F );
use vars qw($originating_system $expiration);

# Parse command line

getopts('a:Ace:f:F:l:m:o:Op:Pq:Rst:v:');
my %data_browse_map;

my $compartment;
if ($opt_f) {
    $compartment = new Safe 'CFG';
    $compartment->share('%cfg_data_browse_map', 
                        '$cfg_msgdir', 
                        '$cfg_pgedir', 
                        '$cfg_logdir', 
                        '$cfg_fail_pge_met_template', 
                        '$cfg_pge_version', 
                        '$cfg_originating_system', 
                        '$cfg_expiration',
                        '%cfg_production_summary',
    );
    $compartment->rdo($opt_f) or
        S4P::perish(30, "main: Failed to read in configuration file $opt_f in safe mode: $!");
    %data_browse_map = %CFG::cfg_data_browse_map;
}

my $msgdir = $opt_m || $CFG::cfg_msgdir || '../MSGS';
my $pgedir = $opt_p || $CFG::cfg_pgedir || '..';
my $logdir = $opt_l || $CFG::cfg_logdir;
my $pge_version = $opt_v || $CFG::cfg_pge_version || basename($opt_p);
my $failpge_met_template = $opt_F || $CFG::cfg_fail_pge_met_template ||
    '../FAILPGE_met_template.cfg';
$originating_system = $opt_o || $CFG::cfg_originating_system || "S4PM";
$expiration = $opt_e || $CFG::cfg_expiration || "3days";

# Setup environment

S4P::logger("INFO", 
    "main: Setting environment: PATH='$pgedir', PGS_PC_INFO_FILE='$ARGV[1]', PGSMSG='$msgdir'..., PGSHOME='$opt_t'");
$ENV{'PATH'} = join(':', $pgedir, $ENV{'PATH'});
$ENV{'PGS_PC_INFO_FILE'} = $ARGV[1];
$ENV{'PGSMSG'} = $msgdir;
$ENV{'PGSHOME'} = $opt_t if ($opt_t);

# Read PCF

my $pcf_name = $ARGV[1];
my $pcf = S4P::PCF::read_pcf($pcf_name) or
              S4P::perish(1, "main: Cannot read/parse PCF $pcf_name: $!");
my ($prefix, $job_type, $job_id, $rest) = split('\.', $pcf_name);
$job_type =~ s/^RUN_//;

# Create symbolic links to input and output directories

S4P::logger("INFO", "main: Setting up local subdirectories and symlinks...");
my (@files, @product_input_files);
my (@support_output_files, @support_input_files);
my (@intermediate_output_files, @intermediate_input_files);

my ($ra_product_output_files, $rh_browse_data_map) =
    get_output_files($pcf->product_output_files, \%data_browse_map, $pcf->product_input_files);
my @product_output_files = @{$ra_product_output_files};
my @product_browse_files = keys %{$rh_browse_data_map};

map {push(@product_input_files, split)} values %{$pcf->product_input_files};
map {push(@support_output_files, split)} values %{$pcf->support_output_files};
map {push(@support_input_files, split)} values %{$pcf->support_input_files};
map {push(@intermediate_output_files, split)} values %{$pcf->intermediate_output_files};
map {push(@intermediate_input_files, split)} values %{$pcf->intermediate_input_files};

# Go through all the referenced files and get their directories

push @files, @product_output_files, @product_browse_files, @product_input_files,
             @support_output_files, @support_input_files,
             @intermediate_output_files, @intermediate_input_files;
my @dirs = unique_dirs(@files);

# Make local directories or symlinks to non-local directories
my @local_dirs = make_local_dirs(@dirs, $opt_s);

my $runlog_file = "$pcf_name.Log";;
$runlog_file =~ s/^DO\.//;
$runlog_file =~ s/^RUN_//;

# Start up production log file...

my $cmd = $ARGV[0];
open RUNLOG, ">$runlog_file" or 
    S4P::perish(113, "main: Cannot open runlog file $runlog_file for writing: $!");
printf RUNLOG << "EOF";
# DPR_ID: $job_type.$job_id
# Resource collection for command:
# $cmd
# Resource Usage Information
EOF
close RUNLOG;
open STDOUT, ">>$runlog_file" or
    S4P::perish(113, "main: Cannot redirect STDOUT to runlog file $runlog_file for appending: $!");

# And we're off and running...

S4P::logger("INFO", "main: Starting up command $cmd...");
my $err_number = S4PM::rusage($cmd);

# Reread PCF in cases where output filenames were tweaked by algorithm
if ($opt_R) {
    S4P::logger('INFO',"Re-reading PCF $pcf_name for new output filenames");
    $pcf = S4P::PCF::read_pcf($pcf_name) or
              S4P::perish(1, "main: Cannot read/parse PCF $pcf_name: $!");
    ($ra_product_output_files, $rh_browse_data_map) =
        get_output_files($pcf->product_output_files, \%data_browse_map, $pcf->product_input_files);
    @product_output_files = @{$ra_product_output_files};
    @product_browse_files = keys %{$rh_browse_data_map};
}

# Move STDOUT back to the real STDOUT so we can close out the RUNLOG file
open STDOUT, ">-" or S4P::perish(113, "main: Cannot redirect STDOUT back to stdout: $!");
# HACK for DPREP exit code 216
if ($err_number == 216) {
    S4P::logger("WARN", "main: Exit 216 from $cmd, setting to 0...");
    $err_number = 0;
}

# If successful, first do the production history
# Then write the output work orders
# Finally clean the log files and temporary stuff
#=====================================================================

my $node_name = $opt_a || S4P::PDR::gethost();

# Generate output UPDATE work order
# N.B.:  Generated for both successes and failures.  In the latter case, it is
# picked up by the "punt" failure handler, as we will no longer be needing the
# input file for this purpose.
# Also, note we are just looking for MODIS/AIRS product input files, not all the
# ancillary files, dynamic or otherwise

# All S4PM-tracked input files have absolute paths
my @input_files = grep(/^\//, @product_input_files);
S4P::perish(130, "main: Cannot find any product input files for UPDATE work order in $ARGV[0]") if (! @input_files);
update_work_order($job_type. '_' . $job_id, \@input_files);

# Finish runlog file no matter what exit code
finish_runlog($runlog_file, $pcf_name, [@product_output_files, @product_browse_files],
    \@support_output_files, \@product_input_files, \@support_input_files);
save_runlog($logdir, $runlog_file) if ($logdir);

#Created Production summary file
if ( %CFG::cfg_production_summary ) {
  unless ( S4PM::QC::production_summary(
					$pcf, 
					\%CFG::cfg_production_summary, 
					$CFG::cfg_originating_system, 
					$runlog_file 
				       )  
	 ) {
      S4P::perish(1, "main: Failed to created production summary file: $!");
    }
}

if ($err_number == 0) {

    # Remove zero-length files (e.g. MOD07_QC) from @product_output_files
    prune_empty_files(\@product_output_files, $job_type . '_' . $job_id);

    if (scalar(@product_output_files) == 0) {
        S4P::logger('WARN', "main: No product output files left after pruning empty ones");
        log2stderr($runlog_file);
    }
    else {
        # Make PDRs for output work orders
        S4P::logger("INFO", "main: Making output work orders...");
        my $catch_pdr = make_pdr('S4P', $node_name, @product_output_files);
        my $export_pdr = make_pdr('ECS', $node_name, @product_output_files);
        $export_pdr->checksum if ($opt_c);

        # Apply Quality Control scripts
        my ($fatal, $ra_qc_clean, $ra_block_export);
        if ($opt_q) {
            ($fatal, $ra_qc_clean, $ra_block_export) = 
                S4PM::QC::apply_qc($opt_q, $export_pdr, $catch_pdr);
            # Write out SWEEP work order if any are files returned by 
            # function
            clean_work_order($job_id, 'QC_BLOCK', @$ra_qc_clean) if (@$ra_qc_clean);
        }
    
        my ($ph_file, $linkage_file);
        if ($opt_P && $export_pdr->total_file_count) {
            # Make the production history
            ($ph_file, $linkage_file) = production_history($pcf_name, $pcf, 0, 
                $export_pdr, $runlog_file);
            S4P::logger("INFO", "main: algorithm successful, production history in $ph_file...");
        }
        else {
            # No production history:  instead cat runlog to STDERR so it ends up in chain log
            log2stderr($runlog_file);
        }

        # Write output PDR work orders
        # Concatenate input job_type with job_id to make output job_id more unique
        # Do a PDR recount() just for good measure...
        $catch_pdr->recount();
        if ($catch_pdr->total_file_count) {
            pdr_work_order('S4P', 'REGISTER', $job_type . $job_id, 
                $node_name, $catch_pdr);
        }
        else {
            S4P::logger('INFO', "No files left for REGISTER PDR after pruning and QC, skipping PDR generation.");
        }
        my $new_job_id = ($opt_O) ? $job_id : ($job_type . '_' . $job_id);
        $export_pdr->recount();
        if ($export_pdr->total_file_count) {
            pdr_work_order('ECS', 'EXPORT', $new_job_id, $node_name, $export_pdr);
        }
        else {
            S4P::logger('INFO', "No files left for EXPORT PDR after pruning and QC, skipping PDR generation.");
        }
        if ($linkage_file) {
            pdr_work_order('ECS', 'EXPORT_PH', $job_type . 'PH_' . $job_id, 
                $node_name, $ph_file, $linkage_file);
        }
        elsif ($opt_P) {
            S4P::logger('WARN', "main: No linkage file, so PH will not be exported.");
        }
        export_browse($rh_browse_data_map, $ra_block_export, 
            $job_type , 'BR_' . $job_id, $node_name);

        # Quit if a fatal error condition was encountered in QC
        S4P::perish(121, "main: Fatal QC error condition encountered.") if $fatal;
    }
    # Cleanup

    S4P::logger("INFO", "main: Cleaning log files and temporary dirs/symlinks...");

    # Collect scratch files, including GetAttr.temp and MCFWrite.temp
    my (@scratch_files);
    my $mcf_temp = $pcf->product_input_files->{'10254.1'};
    my $getattr_temp = $pcf->product_input_files->{'10252.1'};
    push (@scratch_files, $mcf_temp) if $mcf_temp;
    push (@scratch_files, $getattr_temp) if $mcf_temp;
    map {push(@scratch_files, $_)} values %{$pcf->temporary_i_o};
    push @scratch_files, @support_output_files;

    # Clean scratch and temporary files, then temporary directories
    unlink @scratch_files or S4P::logger('WARN', "main: Cannot unlink scratch files " . 
        join(', ', @scratch_files) .  ": $!");

    # Clean those pesky toolkit temporary files:  all pcNNNNNNN... pattern
    opendir DIR, '.';
    map {unlink $_ or S4P::logger('ERROR', "main: Cannot delete $_: $!")} 
        grep(/^pc\d+(\.met)?$/, readdir DIR);
    closedir DIR;

    # Clean temporary directories
    clean_tmp_dirs(@local_dirs) if @local_dirs;
    exit(0);
}
else {
    S4P::logger("ERROR", "main: Algorithm failed with code=$err_number");
    my ($failpge) = production_history($pcf_name, $pcf, $err_number);

    # Use $job_type for pge name in metfile
    failpge_metfile($failpge_met_template, $failpge, $job_type, $pge_version, 
        $job_id, $err_number) or 
        S4P::perish(120, "main: Failed to make metfile for FAILPGE tar file");
    pdr_work_order('ECS', 'EXPORT_FAILPGE', $job_type . 'FAILPGE_' . $job_id, 
        $node_name, $failpge);
    exit($err_number);
}

##########################################################################
# Subroutines
##########################################################################
# clean_tmp_dirs(@dirs)
#   @dirs - list of directories to be removed
#   This unlinks symlinks and recursively removes local subdirectories
#=========================================================================
sub clean_tmp_dirs {
    my ($dir, $file);
    foreach $dir(@_) {

        # Aaack! Don't remove current or parent!
        next if ($dir eq '.' || $dir eq '..');
        next if ($dir =~ m/^\s*$/);
        next if ($dir =~ m/^\//);

        # If it is not just a symlink, clean out the files
        if (! -l $dir) {
            foreach $file(glob("$dir/*")) {
                unlink($file) or 
                    S4P::perish(102, "clean_tmp_dirs(): Cannot unlink local file $file: $!");
            }
        }

        # Remove the symlink or directory
        unlink $dir or 
            S4P::perish(103, "clean_tmp_dirs(): Cannot unlink local directory $dir: $!");
    }
    return 1;
}
# Output SWEEP work order
sub clean_work_order {
    my ($job_id, $clean_type, @files) = @_;
    # Write out SWEEP work order for the actual deletion and deallocation
    my $clean_work_order = "SWEEP.$clean_type" . "_$job_id.wo";
    open SWEEP_WORK_ORDER, ">$clean_work_order" or
        S4P::perish(137, "clean_work_order(): Cannot open $clean_work_order to write: $!");
    foreach my $file(@files) {
        print SWEEP_WORK_ORDER "$file\n";
    }
    close SWEEP_WORK_ORDER;
    return $clean_work_order;
}
sub export_browse {
    my ($rh_browse_data_map, $ra_block_export, $job_type, $job_id, $node_name) = @_;
    my %browse_data_map = %{$rh_browse_data_map} or return;
    my $browse_file;
    my $n_pdr;
    my %block_export = map {($_, 1)} @$ra_block_export;
    # Loop through browse files
    my @export_files;
    foreach $browse_file(keys %browse_data_map) {
        # Handle phantom browse so no failure due to missing browse
        if (! -f $browse_file) {
            S4P::logger('INFO', "clean_work_order(): Browse file $browse_file not generated");
            next;
        }
        # Linkage file is named same as browse file, but with "LINKFILE" as
        # pseudo-data-type, and .pvl at the end
        my ($base, $dir, $ext) = fileparse($browse_file, '\.hdf');
        my $linkage_file_name = "$dir/LINKFILE.$base.pvl";

        # Screen out files that we are not sending
        my @data_files = grep {! exists $block_export{$_}} 
            @{$browse_data_map{$browse_file}};

        # Skip creating linkage or PDR if there are no data files to link to
        unless (@data_files) {
            S4P::logger('INFO', "export_browse(): No data files left after export block to link to $browse_file");
            next;
        }

        # Create the linkage file
        linkage_file($linkage_file_name, @data_files);
        push @export_files, $browse_file, $linkage_file_name;
        S4P::logger('INFO', "export_browse(): Created linkage file $linkage_file_name");

        # Create the PDR:  note one per browse file, because it is easier
    }
    unless ($#export_files < 0) {
        return ( pdr_work_order('ECS', 'EXPORT_BROWSE',
            $job_type . '_' . $job_id, $node_name, @export_files) );
    } else {
        return(1) ;
    } 
}
##############################################################################
# fail_pge_metfile($template, $failpge, $pge, $version, $job_id, $exit_code)
#   $template - metadata template
#   $failpge - Failed algorithm tar file
#   $pge - Name of the pge (for met file)
#   $version - Algorithm Version (for met file)
#   $job_id - Job identifier (for met file - half of DPR_ID)
#   $exit_code - Algorithm exit code (for met file)
#------------------------------------------------------------------------------
#   This creates the metadata file for the FAILPGE.
#==============================================================================
sub failpge_metfile {
    my ($template, $failpge, $pge, $version, $job_id, $exit_code) = @_;
    my $outfile = "$failpge.met";
    if (! $version) {
        S4P::logger('ERROR', "failpge_metfile(): Cannot create FAILPGE metfile without algorithm version (-v): $!");
        return 0;
    }
    if (! open IN, $template) {
        S4P::logger('ERROR', "failpge_metfile(): Cannot open input metfile template $template: $!");
        return 0;
    }
    if (! open OUT, ">$failpge.met") {
        S4P::logger('ERROR', "failpge_metfile(): Cannot open output metfile $failpge.met: $!");
        return 0;
    }
    my $size = (-s $failpge) / (1024. * 1024);
    if (! $size) {
        S4P::logger('ERROR', "failpge_metfile(): $failpge is zero size");
        return 0;
    }
    my $insert_time = S4P::TimeTools::CCSDSa_Now();
    # SSS cannot handle T and Z in times
    $insert_time =~ s/[TZ]/ /g;
    while (<IN>) {
        s/INSERT_SIZE_HERE/$size/ ||
        s/INSERT_PGEVERSION_HERE/$version/ ||
        s/INSERT_PGE_HERE/$pge/ ||
        s/INSERT_EXIT_CODE_HERE/$exit_code/ ||
        s/INSERT_JOB_ID_HERE/$pge.$job_id/ ||
        s/INSERT_TIME_HERE/$insert_time/;
        print OUT;
    }
    close IN;
    close OUT;
    return 1;
}
##############################################################################
# finish_runlog($file, $pcf_file, $ra_product_output, $ra_support_output,
#               $ra_product_input, $ra_support_input)
#   $file - output file for runlog
#   $pcf_file - Process Control File for algorithm run
#   $ra_product_output - reference to array of product output files
#   $ra_support_output - reference to array of support output files (e.g. logs)
#   $ra_product_input - reference to array of product input files
#   $ra_support_input - reference to array of support input files (including
#                       leapsec.dat and utcpole.dat)
#------------------------------------------------------------------------------
#   This finishes off the production log in an effort to look just like
#   the ECS production logs.  It includes:
#   o  Algorithm Completion time
#   o  First line of leapsec.dat and utcpole.dat
#   o  Output filenames
#   o  Input filenames
#==============================================================================
sub finish_runlog {
    my ($file, $pcf_file, $ra_product_output, $ra_support_output, 
        $ra_product_input, $ra_support_input) = @_;
    if (!open LOG, ">>$file") {
        S4P::logger("WARN", "finish_runlog(): Cannot open runlog file $file for append: $!");
        return 0;
    }

    # algorithm Completion time

    print LOG "# Algorithm Completion time (GMT) in Toolkit format\n";
    my ($sec,$min,$hour,$mday,$mon,$year) = gmtime(time);
    $year += 1900;
    printf LOG 
        "PGE_COMPLETION_TIME_TOOLKIT_FORMAT=\"%04d-%02d-%02dT%02d:%02d:%02d\"\n",
        $year, $mon+1, $mday, $hour, $min, $sec;
    print LOG "# Algorithm Completion time (GMT) in understandable format\n";
    print LOG "# Format : Year[Number of years since 1900]:Month[0,11]:Day[1,31]:Hour[0,23]:Min [0,59]:Sec[0,61]\n";
    print LOG "PGE_COMPLETION_TIME_YEAR=$year\n";
    print LOG "PGE_COMPLETION_TIME_MONTH=$mon\n";
    print LOG "PGE_COMPLETION_TIME_DATE=$mday\n";
    print LOG "PGE_COMPLETION_TIME_HOUR=$hour\n";
    print LOG "PGE_COMPLETION_TIME_MIN=$min\n";
    print LOG "PGE_COMPLETION_TIME_SEC=$sec\n\n";

    # First line of leapsec.dat and utcpole

    my @ancillary = grep /(leapsec|utcpole)/, @$ra_support_input;
    if (@ancillary) {
        map {$_ =~ s/^~/$ENV{'PGSHOME'}/} @ancillary if ($ENV{'PGSHOME'});
        print LOG "ANCILLARY_DATA:\n";
        map { printf LOG "%s=%s\n", 
            (/leapsec/) ? 'LEAPSECONDSFILE' : 'ORBITPOLARMOTION', `head -1 $_`}
            @ancillary;
    }

    # Output filenames
    print LOG "File names and actual sizes after algorithm completed\n";
    map {printf LOG "FILE=%s=%d\n", $_, -s $_} @$ra_product_output;

    # Input filenames:  all input filenames with absolute paths
    # This screens out static files (relative paths), toolkit files
    # (~ paths or no paths), temporary files (. paths)
    map {printf LOG "RUNTIME_FILE=%s\n", $_} grep /^\//, @$ra_product_input;
    close LOG;
}

##########################################################################
# get_esdt_version_lgid($pathname)
#   $pathname - pathname of file for which to get LGID
#     if this does not end with .met, it looks in $pathname.met
#-------------------------------------------------------------------------
# LGID specification from Volume 0 of the ECS-SIPS specification:
# LGID:<ESDT>:<ESDTVersion>:<filename>
#=========================================================================
sub get_esdt_version_lgid {
    my $pathname = shift;
    my $metfile = find_metfile($pathname);
    my %attr = S4P::MetFile::get_from_met($metfile, 'SHORTNAME', 'VERSIONID', 
        'LOCALGRANULEID');
    S4P::perish(91, "get_esdt_version_lgid(): Cannot find SHORTNAME in $metfile") if !$attr{'SHORTNAME'};
    S4P::perish(92, "get_esdt_version_lgid(): Cannot find VERSIONID in $metfile") if !$attr{'VERSIONID'};
    my $lgid = sprintf("LGID:%s:%03d:%s", $attr{'SHORTNAME'}, 
        $attr{'VERSIONID'}, $attr{'LOCALGRANULEID'}) if $attr{'LOCALGRANULEID'};
    return ( $attr{'SHORTNAME'}, $attr{'VERSIONID'}, $lgid);
}
##############################################################################
# linkage_file($pathname, @files)
#   $pathname - output pathname for linkage file
#   @files - data files to point to
# ----------------------------------------------------------------------------
# This creates a linkage file for linking production histories and browse
# with the output granules they refer to.  The link is done via Local
# Granule ID, as per Volume 0 of the ECS-SIPS ICD.
#=============================================================================
sub linkage_file {
    my ($pathname, @files) = @_;
    my $file;
    my $n_files = 0;
    my ($file, @linkage);
    # Loop through files, using only those which:
    # have non-zero size and LocalGranuleID
    foreach $file(@files) {
        # Filter out zero-length files
        next unless (-s $file);
        my ($esdt, $version, $lgid) = get_esdt_version_lgid($file);
        # Filter out those which have no lgid
        if ($lgid) {
            $n_files++;
            push(@linkage, sprintf("  GRANULE_POINTER_%03d=\"%s\"", $n_files,$lgid));
        }
    }
    if (! $n_files) {
        S4P::logger('WARN', "linkage_file(): No files to link in linkage file $pathname");
        return;
    }
    open LINKAGE, ">$pathname" or 
        S4P::perish(123, "linkage_file(): Cannot open linkage file $pathname to write: $!");
    print LINKAGE "GROUP=LINKAGE_POINTERS\n";
    printf LINKAGE "  NUM_POINTERS=%d\n", $n_files;
    print LINKAGE join("\n", @linkage);
    print LINKAGE "\nEND_GROUP=LINKAGE_POINTERS\nEND\n";
    close(LINKAGE);
    return $pathname;
}
##########################################################################
# log2stderr($runlog_file)
#   $runlog_file - pathname of run-log file
# Cats the run-log file to STDERR and deletes it
##########################################################################
sub log2stderr {
    my $runlog_file = shift;
    open RUNLOG, $runlog_file;
    while (<RUNLOG>) {
        print STDERR;
    }
    close RUNLOG;
    unlink $runlog_file;
}
##########################################################################
# make_local_dirs(@dirs)
#   @dirs - list of local directories to either symlink or make
# ----------------------------------------------------------------------------
# This makes local subdirectories for paths that start with './' and
# symlinks for those that do not.
# It is used to make all of the input and output directories for a algorithm run
# available "locally", within the job directory, for easy troubleshooting.
#=============================================================================
sub make_local_dirs {
    my ($dir, $use_symlinks);
    my (@dirs, $link);
    foreach $dir(@_) {
        # Skip if it's just the current or parent directory
        next if ($dir eq '.' || $dir eq '..');

        # Make a subdirectory if it starts with ./
        if ($dir =~ m#^\./#) {
            mkdir($dir, 0775) or 
                S4P::perish(2, "make_local_dirs(): Cannot make local dir $dir: $!");
            push @dirs, $dir;
        }

        # Otherwise, create a symbolic link to the directory
        elsif ($use_symlinks) {
            $link = $dir;
            $link =~ s#\.\.#+#g;
            $link =~ s#/#_#g;
            $link =~ s/^_//;
            # Screen out circular links and non-existent directories
            # We don't warn of these, as they can be common for unused LUNs
            # in a pcf.
            next if ($dir eq $link);
            next if (! -d $dir);
            symlink($dir, $link) or 
                S4P::logger('WARN', "make_local_dirs(): Cannot create symlink $link to $dir: $!");
            push @dirs, $link;
        }
    }
    return @dirs;
}
##########################################################################
#=============================================================================
sub make_pdr {
    my ($pdr_type, $node_name, @files) = @_;
    $expiration =~ m/(\d+)\s*(\w+)/;
    my $exp_number = $1;
    my $exp_units = $2;
    my $pdr = S4P::PDR::start_pdr('originating_system' => $main::originating_system,
        'expiration_time' => S4P::PDR::get_exp_time($exp_number, $exp_units));

    # Loop through each filename in the PDR
    my ($i, $n_files, $penultimate);
    $n_files = scalar(@files);
    $penultimate = $n_files - 1; 
    for ($i = 0; $i < $n_files; $i++) {
        my ($filename, undef) = fileparse($files[$i]);
        my $granule;

        # Production history must be followed by its linkage file in the arglist
        # Regarless of S4PM file name pattern, PH files will always be named
        # PH.*

        my $esdt;
        if ( $filename =~ /^PH/ ) {
            $esdt = 'PH';
        } else {
            ($esdt, undef, undef, undef, undef) =
                S4PM::parse_patterned_filename($files[$i]);
        }
        S4P::logger("INFO", "make_pdr(): esdt: [$esdt]");

        if ($i < $penultimate && ($files[$i+1] =~ /LINKFILE\..*pvl$/)) {
            my $version = '001';
            # Will have a .met if this is Browse
            ($esdt, $version) = get_esdt_version_lgid(find_metfile($files[$i]))
                unless ($filename =~ /^PH/);
            $granule = $pdr->add_granule(
                'data_type' => $esdt, 
                'data_version' => $version,
                'node_name' => $node_name,
                'files'=>[$files[$i], $files[$i+1]]);
            if ($filename !~ /^PH/) {
                $granule->file_specs->[0]->file_type('BROWSE');
                $granule->data_type('Browse');
            }
            $i++;
        }
        else {
            my $metfile = find_metfile($files[$i]);
            my ($esdt, $version, $lgid) = get_esdt_version_lgid($metfile);
            $granule = $pdr->add_granule(
                'data_type' => $esdt,
                'data_version' => $version,
                'node_name' => $node_name,
                'files'=>[$files[$i], $metfile]);

            # Add LGID as UR if the output PDR is a non-ECS-standard one
            # This is so it can be saved in the database and later used for
            # InputGranulePointer

            $granule->ur($lgid) if ($pdr_type ne 'ECS');
        }
    }
    $pdr->resolve_paths;
    return $pdr;
}
############################################################################
# Find metadata file
############################################################################
sub find_metfile {
    my $data_file = shift;

    # (1) Look for file passed in as argument if it has a .met or .xml
    # (2) If argument is a data file, first try adding .met
    my $met_file = ($data_file =~ /\.(met|xml)/) ? $data_file : "$data_file.met";
    return $met_file if (-f $met_file);

    # (3) If no .met file exists, then try adding .xml
    $met_file =~ s/\.met$/.xml/;
    return $met_file if (-f $met_file);

    # (4) All attempts failed
    warn "Cannot find either metadata file $data_file.met or $data_file.xml";
    return;
}

sub get_output_files {
    my ($rh_product_output_files, $rh_data_browse_map, $rh_product_input_files) = @_;
    my %data_browse_map = %{$rh_data_browse_map};
    my @product_output_files;

    # Configuration map is based on luns, with data lun as key and browse lun
    # as value.  There can be only one browse file for a given data file
    # Browse-data map reverses this pattern, using files:
    # Key is pathname of browse file, value is anonymous array of data files
    my %browse_data_map;
    my %output = %{$rh_product_output_files};
    my %input = %{$rh_product_input_files};
    my ($data_lun, $browse_lun, %delete_browse_lun);

    # Loop through each browse lun in the browse map
    while ( ($data_lun, $browse_lun) = each %data_browse_map) {
        # No need to split browse files in that hash value, can only be 1
        my $browse_file = $output{$browse_lun};
        my @data_files = split('\s', $output{$data_lun});
    
        # Add array of data files for this LUN to the hash
        push @{$browse_data_map{$browse_file}}, @data_files;

        # Mark browse for delete from product output
        # browse files are kept separate so they go into separate EXPORT
        # work order
        # However, don't delete yet, as we may need it in later iterations
        # of this while loop
        $delete_browse_lun{$browse_lun} = 1;

        # Now check input files for possible links
        # This allows linking to output files from upstream algorithms
        if ($input{$data_lun}) {
            @data_files = split('\s', $input{$data_lun});
            push @{$browse_data_map{$browse_file}}, @data_files;
        }
    }
    map {delete $output{$_}} keys %delete_browse_lun;
    map {push(@product_output_files, split)} values %output;
    return (\@product_output_files, \%browse_data_map);
}

##########################################################################
# pdr_work_order($pdr_type, $job_type, $job_id, $node_name, @files)
# ----------------------------------------------------------------------------
#   Creates a PDR work order (actually most of the work is done in make_pdr()).
#=============================================================================
sub pdr_work_order {
    my ($pdr_type, $job_type, $job_id, $node_name, @files) = @_;

    my $filename = "$job_type.$job_id.wo";
    my $pdr = ref($files[0]) ? $files[0] : make_pdr($pdr_type, $node_name, @files);
    S4P::perish(110, "pdr_work_order(): Failed to make PDR for output work order $filename") 
        unless $pdr;

    # pdr->write_pdr has weird error return:  non-zero if failure
    if ($pdr->errors) {
        S4P::perish(111, "pdr_work_order(): Errors in PDR for output work order $filename");
    }
    elsif ($pdr->write_pdr($filename)) {
        S4P::perish(112, "pdr_work_order(): Failed to write output work order $filename");
    }
}

##########################################################################
# production_history($pcf_name, $pcf, $err_number, $export_pdr, @files) 
#   $pcf_name - name of Process Control File
#   $pcf - parsed PCF object
#   $err_number - exit code of algorithm
#   @files - all files to be included in production history (more will be added)
# ----------------------------------------------------------------------------
# Makes both Production History and FAILPGE tar files, depending on what 
# the error code was.
# The two are very similar, the main difference (aside from the ESDT and name)
# being whether a linkage file is made to point to the output data granules.
# (FAILPGE has no linkage file.)
#=============================================================================
sub production_history {
    my ($pcf_name, $pcf, $err_number, $export_pdr, @files) = @_;

    # N.B.:  Filename is PH.[AP]AYYYYDDDHHMM.VVV.YYYYDDDHHMMSS.tar
    # Linkage file is LINKFILE.[AP]AYYYYDDDHHMM.VVV.YYYYDDDHHMMSS.pvl
    # .Log includes rusage info, LEAPSECONDSFILE, ORBITPOLARMOTION
    # .Tk{User,Report,Status} = Log{User,Report,Status} file

    my ($prefix, $root) = split('\.', $pcf_name, 2);
    my $chain_log = "$root.log";
    if (-f $chain_log) {
        my $logname = "$chain_log.chain";
        copy $chain_log, $logname;
        push(@files, $logname) if (-f $logname);
    }
    $root =~ s/^RUN_//;

    # Add symlinks to log files
    my %log_files = %{ $pcf->log_files };
    my ($err, $ftype);
    foreach $ftype('status', 'report', 'user') {
        if (! -f $log_files{$ftype}) {
            $err++;
            S4P::logger('WARN', 
                "production_history(): Cannot find $ftype logfile $log_files{$ftype}");
        }
        else {
            my $suffix = ucfirst $ftype;
            my $filename = "$root.Tk$suffix";
            symlink ($log_files{$ftype}, $filename) if ($filename ne $log_files{$ftype}) ;
            push @files, $filename;
        }
    }

    # Add core file if there
    if ($err_number && -f "core") {
        rename ("core", "$root.core");
        push (@files, "$root.core");
    }

    # Add PCF
    symlink ($pcf_name, "$root.pcf");
    push (@files, "$root.pcf");

    # Create tar file, following symbolic links
    my ($type, $tarfile, $linkage_file);

    # If it failed, make a FAILPGE package
    if ($err_number) {
        $tarfile = "FAILPGE.$root.tar";
    }

    # Successful Algorithm:  will make a production history and linkage file
    else {
        my @output_files = $export_pdr->files('SCIENCE');

        # The PH directory is normally in the same tree as the output files,
        # but in a subdirectory named PH
        my $dir = dirname($output_files[0]) if ($output_files[0]);
        my $ph_dir = ($dir) ? (dirname($dir) . '/PH') : '../../DATA/PH';

        # But if that subdirectory doesn't exist, put it in the same directory
        # as the first output file you come to
        if (! -d $ph_dir) {
            S4P::logger('WARN', "production_history(): Normal production history directory $ph_dir not found; using $dir instead");
            $ph_dir = $dir;
        }
        $tarfile = "$ph_dir/PH.$root.tar";
        $linkage_file = linkage_file("$ph_dir/LINKFILE.$root.pvl", @output_files);
    }

    # Tar up the appropriate files in the directory
    my $osname = `/bin/uname -s`;
    my $tar_args = ($osname =~ /IRIX64/) ? "-Lcf" : "-hcf";
    my ($errstr, $errnum) = S4P::exec_system('tar', $tar_args, $tarfile, @files);
    if ($errnum) {
        S4P::logger("ERROR", "production_history(): Cannot make tar file $tarfile: $errstr");
        return undef;
    }

    # Remove all of the symlinks and the core
    unlink @files;
    return ($tarfile, $linkage_file);
}

##########################################################################
# prune_empty_files(\@product_output_files, $job_id)
#   \@product_output_files - pointer to product_output_files
#----------------------------------------------------------------------------
# This generates a SWEEP work order for clean_data so it can be 
# deallocated.  It also removes them from the product_output_files
# array so that they don't show up in PDRs.
#=============================================================================
sub prune_empty_files {
    my ($ra_product_output, $job_id) = @_;
    my ($i, $n, $file);
    $n = scalar(@{$ra_product_output});
    my @empty_files = ();
    # Loop through all files in array
    # in reverse order so as not to clobber indices with splice
    for ($i = $n-1; $i >= 0; $i--) {
        $file = $ra_product_output->[$i];
        # If zero-length or non-existent, then delete (unlink) and 
        # remove from array (splice)
        if (! -s $file) {
            $file =~ s/\n$//;  # Try to eliminate mystery blank line
            push (@empty_files, $file) if $file;
            splice(@{$ra_product_output}, $i, 1);
        }
    }
    return "No empty files" if (! @empty_files);
    return clean_work_order($job_id, 'EMPTY', @empty_files);
}
sub save_runlog {
    my ($dir, $file) = @_;
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday) = localtime(time);
    my $dest = sprintf("%s/%04d%03d_%02d", $dir, $year+1900, $yday+1, $hour);
    if (! -d $dest && !mkdir($dest, 0755)) {
        S4P::logger('WARN', "save_runlog(): Failed to create log directory $dest: $!");
        return 0;
    }
    if (!copy($file, "$dest/$file")) {
        S4P::logger('WARN', "save_runlog(): Failed to copy runlog $file to $dest: $!");
        return 0;
    }
    return 1;
}
##########################################################################
# unique_dirs(@pathnames)
#   @pathnames - list of pathnames
#----------------------------------------------------------------------------
# Outputs a unique list of the directories holding the pathnames in the input
# arguments.
#=============================================================================
sub unique_dirs {
    my %dirs;

    # Replace ~ with PGSHOME if it is set
    if ($ENV{'PGSHOME'}) {
        map {$_ =~ s/^~/$ENV{'PGSHOME'}/} @_;
    }
    # Put all the directories in a hash, then return the keys
    map { $dirs{dirname($_)} = 1; } @_;
    return (keys %dirs);
}

##########################################################################
# update_work_order($job_id, $ra_input_files)
#   $job_id - job identifier
#   $ra_input_files - reference to an array of input files
#----------------------------------------------------------------------------
# Creates the UPDATE work order, decrementing the Uses for each input file.
#=============================================================================
sub update_work_order {
    my ($job_id, $ra_input_files) = @_;

    # Open UPDATE work order output file
    my $filename = "UPDATE.UPDATE_" . "$job_id.$$.wo";
    open (UPDATE, ">$filename") or 
        S4P::perish(120, "update_work_order(): Cannot open output work order $filename");

    # Multi-file entities have to be stripped back to their dirname

    # Update uses for files
    my %granule_ids;
    foreach my $file( @{$ra_input_files}) {
        # Save multi-file guys for later
        if ( ($file =~ /RM[AT]000\./) || ($file =~ /PDS$/) ) {
            $granule_ids{dirname($file)} = 1;
        }
        else {
            printf UPDATE "FileId=%s Uses=-1\n", $file;
        }
    }
    # Multi-file granules (MODIS L0, AIRS L0...)
    foreach my $granule(keys %granule_ids) {
        printf UPDATE "FileId=%s Uses=-1\n", $granule;
    }
    close(UPDATE);

    return $filename;
}
