#!/usr/bin/perl

=head1 NAME

s4pm_show_pge.pl - pull version string out of PCF templates to be picked up by 
ShowPGE

=head1 SYNOPSIS

s4pm_show_pge.pl

=head1 DESCRIPTION

B<s4pm_show_pge.pl> is meant to be run as a pseudo-cron that retrieves the 
GDAAC version string from a algorithm's PCF template (in LUN 411) and places 
it in a version file located in the S4PM user's home directory. The ShowPGE 
tool (ShowPGE(1)) queries all version files when run by a user.

The version files are placed in $HOME/<machine>/<instance>/<gear>/<PGEname_ver>
and are named 'version_I<version>' where I<version> is replaced by the actual
version as set in the Run Algorithm station.cfg file. For example: 
version_3.2.0.  B<s4pm_show_pge.pl> assumes that there is a subdirectory 
under $HOME for each combination of box, instance, and gear of S4PM running. 

Within each S4PM instance directory, B<s4pm_show_pge.pl> will create (if 
necessary) a subdirectory for each algorithm name currently running within 
that S4PM instance. The information as to what algorithm name and version is 
currently running is pulled out of the active station.cfg file in the 
Run Algorithm station.

If anything already exists in the algorithm directory (including a previously
generated 'version' file), it is deleted. A new file is created each time
that contains only a single line. This line is the version string pulled
out of the PCF template file.

=head1 AUTHOR

Stephen W Berrick, NASA/GSFC, Code 610.2

=cut

################################################################################
# s4pm_show_pge.pl,v 1.2 2006/09/12 20:31:39 sberrick Exp
# -@@@ S4PM, Version Release-5_16_0
################################################################################

use strict;
use S4P;
use S4PM;
use Cwd;
use Safe;

my $pgename;
my $version;
my $template;
my $pcf_template_path;
my $instance;
my $version_string;

# Figure out what S4PM instance I am based on my current directory

my($machine, $mode, $instance, $gear) = S4PM::get_box_mode_instance_gear();

if ( $machine eq 'ERROR' ) {
    S4P::perish(100, "main: Could not parse current directory properly.");
}

# Read the run_algorithm station.cfg file to see what algorithms and algorithm 
# versions are now running

my $sconfig = "../../run_algorithm/station.cfg";
my $compartment = new Safe 'CFG';
$compartment->share('%cfg_commands');
$compartment->rdo($sconfig) or 
    S4P::perish(30, "main: Failed to read in configuration file $sconfig in safe mode: $!");

foreach my $key ( keys %CFG::cfg_commands ) {
    $pgename = $key;
    $pgename =~ s/^RUN_//;
    my $cmdstr = $CFG::cfg_commands{$key};
    if ( $cmdstr =~ /^.+\s+-v\s+([^\s]+)\s+.*$/ ) {
        $version = $1;
    } else {
        S4P::perish(40, "main: Could not parse command string: [$cmdstr] for a algorithm version.");
    }
    
    $pcf_template_path = "../../run_algorithm/$pgename/$version/pge";

    my @templates = glob("$pcf_template_path/*.tpl");

    foreach $template (@templates) {
        my $syscmd = "/bin/grep \"^411\" $template";

        my @matches = `$syscmd`;
        unless ( @matches ) {
            $version_string = "No version string defined";
        } else {
            foreach (@matches) {
                chomp;
                $version_string = $_;
                $version_string =~ s/^411\s*\|\s*[^|]+\s*\|//;
            }
        }

        my $home_dir = $ENV{'HOME'};
        my $dir_name = "$home_dir/$machine/$instance/$gear/show_pge";

####### Check for existence of the correct directory for this instance of S4PM
####### in the home directory. Perish if it's not there.

        if ( ! -e "$dir_name" ) {
            S4P::perish(80, "main: Directory [$dir_name] doesn't seem to exist!");
        }
        
####### Check for existence of the directory for this algorithm. If it's not 
####### there, create it. 

        if ( ! -e "$dir_name/$pgename" ) {
            S4P::logger("INFO", "main: Directory $dir_name/$pgename doesn't exist. I'll create it.");
            mkdir("$dir_name/$pgename", 0755) or 
                S4P::perish(70, "main: Cannot mkdir $dir_name/$pgename with 0755: $!");
        } else {
            system("/bin/rm -fr $dir_name/$pgename/*");
        }
            
####### Finally, create the version file in this directory.

        open(VER, ">$dir_name/$pgename/version_$version") or 
            S4P::perish(110, "main: Could not open file for writing: [$dir_name/$pgename/version_$version]: $!");
        print VER $version_string;
        close(VER) or S4P::perish(110, "main: Could not close file from writing: $dir_name/$pgename/version_$version: $!");
    }
    
}

exit 0;

