#!/tools/gdaac/COTS/perl-5.8.5/bin/perl

=head1 NAME

s4pm_sweep_data.pl - removes files no longer needed & deallocates resources

=head1 SYNOPSIS

s4pm_sweep_data.pl 
B<[-a]>
[B<-d[b]> I<allocate_disk_db>]
[B<-l[ink]>]
B<-c[onfig]> I<allocate_disk config>
B<-e[xpire]> I<expiration_file>
I<SWEEP Workorder>

=head1 DESCRIPTION

This program is based upon DMcleanup.pl.

B<s4pm_sweep_data.pl> reads the input SWEEP work order and deletes the 
files contained therein. The SWEEP work order file name is assumed to be:
DO.SWEEP.I<jobid>.wo and this file is assumed to only contain a list of 
full pathnames of files to delete, one per line. 

From the input SWEEP work order, B<s4pm_sweep_data.pl> will delete each 
granules. For multi-file granules (such as MOD000), B<s4pm_sweep_data.pl> will 
first delete the individual files and then the directory containing those 
files.

After successful deletion of each granule, the disk resource for that granule
is deallocated from the disk pool. This is done via a call to the ResPool
module.

For those granules associated with a Request Data station stub file, the 
associated stub file is deleted as well.

No output work order is produced. 

B<s4pm_sweep_data.pl> uses the Allocate Disk configuration file and the 
Allocate Disk database file, both of which are assumed to be in the station 
directory (or, at least links to them).

=head1 ARGUMENTS

=over 4

=item B<-a> or B<-actual>

Use actual sizes in allocation. In this mode, the actual size of the file is 
deallocated instead of the maximum size specified in the configuration file.

=item B<-d[b]> I<allocate_disk_db>

Database (DBM) file to deallocate from. This option is required if
B<-actual> is set. It is otherwise ignored.

=item B<-l[ink]>

Causes no deallocation to be performed for files that are really just symbolic 
links, presumably because these files are never allocated space in the first
place (see the -l option of s4pm_request_data.pl).

=item B<-c[onfig]> I<allocate_disk config>

Configuration file with allocation mappings and maximum sizes.

=item B<-e[xpire]> I<expiration_file>

File with granules that are ripe for expiration, as they are all used up.
Generally speaking, this will be just files in the INPUT pool. (Yes, it's a 
"magic string"; we'll override it later if necessary. For now, YAGNI.)
The program gets the URs from the request_data/REQUESTS directory and writes
them to this file.

=back

=head1 AUTHORS

Bob Mack, NASA/GSFC, Code 902, Greenbelt, MD 20771

Stephen Berrick, NASA/GSFC, Code 610.2

Christopher Lynnes, NASA/GSFC, Code 610.2

=cut

################################################################################
# s4pm_sweep_data.pl,v 1.2 2006/09/12 20:31:39 sberrick Exp
# -@@@ S4PM, Version Release-5_16_0
################################################################################

use strict;
use S4P;
use S4PM;
use Getopt::Long;
use S4P::ResPool;
use DB_File;
use File::Basename;
use Safe;

################################################################################
# Global variables                                                             #
################################################################################
 
use vars qw($InputWorkorder
            $ConfigFile
            $ALLOCDB
            $req_stub_dir
            $use_actual_size
            $obey_links
            $expiration_file
);
 
# Relative directory where the request stub files reside

$req_stub_dir = "../../request_data/REQUESTS";

$ALLOCDB = undef;

################################################################################

# Process command-line arguments

GetOptions( "config=s" => \$ConfigFile,
            "db=s"     => \$ALLOCDB,
            "actual"   => \$use_actual_size,
            "link"     => \$obey_links,
            "expire=s" => \$expiration_file,
 );

unless ( $ConfigFile ) {
    S4P::perish(10, "main: No Allocate Disk configuration file was specified on the command line with the -c argument.");
}
unless ( -e $ConfigFile ) {
    S4P::perish(30, "main: Allocate Disk configuration file: $ConfigFile doesn't seem to exist!");
}
unless ( $ALLOCDB ) {
    S4P::perish(10, "main: No Allocate Disk database file was specified on the command line with the -d argument.");
}
unless ( -e $ALLOCDB ) {
    S4P::perish(30, "main: The Allocate Disk database file: $ALLOCDB doesn't seem to exist!");
}
$use_actual_size = 1 if ($use_actual_size);  # Make sure value is 1, not some other non-zero value

my $compartment = new Safe 'CFG';
$compartment->share('%datatype_pool_map','%datatype_pool', '%proxy_esdt_map', '%datatype_maxsize');
$compartment->rdo($ConfigFile) or 
    S4P::perish(1, "main: Cannot read config file $ConfigFile in safe mode: $!");
my $rh_pool = \%CFG::datatype_pool;
my $rh_pool_map = \%CFG::datatype_pool_map;
my $rh_proxy = \%CFG::proxy_esdt_map;
my $rh_maxsize = \%CFG::datatype_maxsize;
 
# Verify that an input work order (last command-line argument) has been
# specified and that it exists. Note that GetOptions (above) will remove
# from ARGV only those arguments it recognizes and leave the remaining in
# place. Since the input work order is not "recognized", it is guaranteed
# to be the only remaining argument and will thus be the first: $ARGV[0]

if ( !defined $ARGV[0] ) {
    S4P::perish(20, "main: No input work order specified!");
} else {
    $InputWorkorder = $ARGV[0];
    unless ( -e $InputWorkorder ) {
        S4P::perish(20, "main: Workorder $InputWorkorder doesn't seem to exist!");
    }
}
 
S4P::logger("INFO", "********** s4pm_sweep_data.pl starting for work order $InputWorkorder **********");

# Read in the input SWEEP work order

my $wo_string = S4P::read_file($InputWorkorder);
unless ( $wo_string ) {
     S4P::perish(20, "main: S4P::read_file: No string read from input work order: $InputWorkorder.");
}
 
### Parse work order string for file pathname

my @files = split(/\n/, $wo_string);

foreach my $file ( @files ) {
    next if ( $file eq '' );
    my ($fn, $dir) = fileparse($file);
    next if ( $file =~ /^\s*$/ );     # Skip if line is blank
    if ( $fn eq "" or ! $fn or $dir eq "" or ! $dir ) {
        S4P::perish(20, "main: Bad or corrupted input work order. Expected full pathnames. Got this instead: [$file]. Something ain't right.");
    }
    S4P::logger("DEBUG", "main: file: [$file]");

### Try getting data type from the UR file, but use file name as a backup:

    my ($datatype, undef, undef, undef, undef) = S4PM::parse_patterned_filename($fn);
    my $urdt;
    if ( -e "$file.ur" ) {     # Prefer getting data type from UR file
        S4P::logger("DEBUG", "main: UR file found.");
        ($urdt, undef, undef) = S4PM::parse_ur_file("$file.ur");
    }
    if ( $urdt ) { $datatype = $urdt; }
    unless ( $datatype or $urdt ) {
        S4P::perish(30, "main: Failed to get data type either from UR file: [$file.ur] or from file name: [$fn].");
    }

    unless ( -e $file ) {
        S4P::logger("WARNING", "main: File: [$file] does not seem to exist! Freeing the disk space anyway.");

####### We go ahead and free the disk space even if the file doesn't exist

        unless ( S4PM::free_disk($file, $datatype, $ALLOCDB, 
                       $rh_pool, $rh_pool_map, $rh_proxy, $rh_maxsize, 0) ) {
            S4P::perish(60, "main: Failed to de-allocate disk for $file");
        }
        next;
    }
    if ( -d $file ) {	# Multi-file granule
        S4P::logger("DEBUG", "main: file: [$file] is a directory. Calling wipe_directory().");
        my $actual_size = 0;
        foreach my $f(glob("$file/*")) {
            $actual_size += (-f $f);
        }
        unless ( wipe_directory($file, $datatype, $expiration_file) ) {
            if ( $obey_links ) {
                if ( ! -l $file ) {
                    unless ( S4PM::free_disk($file, $datatype, $ALLOCDB, $rh_pool, 
                        $rh_pool_map, $rh_proxy, $rh_maxsize, $use_actual_size, 
                        $actual_size) ) {
                        S4P::perish(60, "main: Failed to de-allocate disk for $file");
                    }
                } else {
                    S4P::logger("INFO", "main: File $file is only a symbolic link and -l option is set. Therefore, no re-allocation being performed.");
                }
            } else {
                unless ( S4PM::free_disk($file, $datatype, $ALLOCDB, $rh_pool, 
                    $rh_pool_map, $rh_proxy, $rh_maxsize, $use_actual_size, 
                    $actual_size) ) {
                    S4P::perish(60, "main: Failed to de-allocate disk for $file");
                }
            }
        } else {
            S4P::logger("DEBUG", "main: Directory: [$file] is NOT being de-allocated.");
        }
    } else {	# Single file granule
        S4P::logger("DEBUG", "main: file: [$file] is a file. Calling wipe_file().");
        # Capture actual size before we delete it!
        my $actual_size = (-s $file);
        unless ( wipe_file($file, $datatype, $expiration_file) ) {
            if ( $obey_links ) {
                if ( ! -l $file ) {
                    unless ( S4PM::free_disk($file, $datatype, $ALLOCDB, $rh_pool, 
                        $rh_pool_map, $rh_proxy, $rh_maxsize, $use_actual_size, 
                        $actual_size) ) {
                        S4P::perish(60, "main: Failed to de-allocate disk for $file");
                    }
                } else {
                    S4P::logger("INFO", "main: File $file is only a symbolic link and -l option is set. Therefore, no re-allocation being performed.");
                }
            } else {
                unless ( S4PM::free_disk($file, $datatype, $ALLOCDB, $rh_pool, 
                    $rh_pool_map, $rh_proxy, $rh_maxsize, $use_actual_size, 
                    $actual_size) ) {
                    S4P::perish(60, "main: Failed to de-allocate disk for $file");
                }
            }
        } else {
            S4P::logger("DEBUG", "main: File: [$file] is NOT being de-allocated.");
        }
    }
}

S4P::logger("INFO", "********** s4pm_sweep_data.pl completed successfully! **********");

exit 0;

sub compare_urs {

################################################################################
#                                  compare_urs                                 #
################################################################################
# PURPOSE: To compare two URs to see if they are equal                         #
################################################################################
# DESCRIPTION: compare_ur compares two URs by comparing the last portion of    #
#              each UR. This last portion contains the ESDT ShortName,         #
#              VersionID, and database ID. To qualify as being equal, this     #
#              last portion must match. The front part of the UR may not match #
#              due to ':DSSDSRV' being replaced by 'GSF:DSSDSRV' or numbers    #
#              being different.                                                #
################################################################################
# RETURN: 1 - URs are equal                                                    #
#         0 - URs are NOT equal                                                #
################################################################################
# CALLS: S4P::logger                                                           #
#        S4P::perish                                                           #
################################################################################
# CALLED BY: wipe_request                                                      #
################################################################################

    my($file_ur, $req_ur) = @_;

    my $file_dbid = "";
    my $req_dbid;

    S4P::logger("DEBUG", "compare_urs(): File UR: [$file_ur]");
    S4P::logger("DEBUG", "compare_urs(): Request UR: [$req_ur]");

### If URs are URLs (ftp or http), we compare directly. But if they are ECS
### style URs, we can't since request stub files contain [:DSSDSRV] while URs
### in .ur files contain [GSF:DSSDSRV] (GSF for Goddard DAAC, in this case).
### Thus, for ECS style URs, we focus specifically on the database IDs which
### will match if the two are equivalent. 

### There's probably a better way to do this.

    
    if ( $file_ur =~ /^ftp/ or $file_ur =~ /^http/ ) {
        $file_dbid = $file_ur;
    } elsif ( $file_ur =~ /^.*\:(.*\.[0-9][0-9][0-9]\:.*)$/ ) {
        $file_dbid = $1;
        $file_dbid =~ s/^\s+//;
        $file_dbid =~ s/\s+$//;
        S4P::logger("DEBUG", "compare_urs(): File DBid: [$file_dbid]");
    }

    if ( $req_ur =~ /^ftp/ or $req_ur =~ /^http/ ) {
        $req_dbid = $req_ur;
    } elsif ( $req_ur =~ /^.*\:(.*\.[0-9][0-9][0-9]\:.*)$/ ) {
        $req_dbid = $1;
        $req_dbid =~ s/^\s+//;
        $req_dbid =~ s/\s+$//;
        S4P::logger("DEBUG", "compare_urs(): Request DBid: [$req_dbid]");
    } else {
        S4P::perish(40, "compare_urs(): Could not parse file UR: $req_ur");
    }

    if ( $file_dbid eq $req_dbid ) {
        S4P::logger("DEBUG", "compare_urs(): Request and File URs are equal");
        return 1;
    } else {
        S4P::logger("DEBUG", "compare_urs(): Request and File URs are NOT equal");
        return 0;
    }
}

sub wipe_file {

################################################################################
#                                  wipe_file                                   #
################################################################################
# PURPOSE: To delete a file                                                    #
################################################################################
# DESCRIPTION: wipe_file deletes the file passed to it and also increments the #
#              success counter (or the failure counter if the delete fails).   #
#              Only when the UR file is deleted is the Request Data station    #
#              request stub file deleted.                                      #
################################################################################
# RETURN: 0 - Successful                                                       #
#         1 - Failure                                                          #
################################################################################
# CALLS: S4P::logger                                                           #
#        wipe_request                                                          #
################################################################################
# CALLED BY: main                                                              #
################################################################################

    my ($file, $datatype, $expiration_file) = @_;

    S4P::logger("DEBUG", "wipe_file(): Entering wipe_file() with file: [$file] and datatype: [$datatype]");

    if ( unlink($file) ) {
        S4P::logger("DEBUG", "wipe_file(): $file deleted."); 
            if ( -e "$file.met" ) {
                unlink("$file.met") or 
                    S4P::logger("WARNING", "wipe_file(): Failed to delete associated ODL metadata file: $file.met.");
            }
            if ( -e "$file.xml" ) {
                unlink("$file.xml") or 
                    S4P::logger("WARNING", "wipe_file(): Failed to delete associated XML metadata file: $file.xml.");
            }
            if ( -e "$file.ur" ) {
                my $ur = wipe_request("$file.ur", $datatype, $expiration_file);
                unlink("$file.ur") or 
                    S4P::logger("WARNING", "wipe_file(): Failed to delete associated UR file: $file.ur.");
            }
    } else {
        S4P::logger("WARNING", "wipe_file(): Failed to delete $file.");
        S4P::logger("DEBUG", "wipe_file(): Leaving wipe_file()");
        return 1;
    }
 
    S4P::logger("DEBUG", "wipe_file(): Leaving wipe_file()");
    return 0;
}

sub wipe_directory {

################################################################################
#                               wipe_directory                                 #
################################################################################
# PURPOSE: To delete a directory and its contents                              #
################################################################################
# DESCRIPTION: wipe_directory deletes a directory and its contents. It also    #
#              increments the success counter (or the failure counter if the   #
#              delete fails). The deletion of the directory and its contents   #
#              is considered a single deletion. Only when the UR file is       #
#              deleted is the Request Data station request stub file deleted.  #
################################################################################
# RETURN: 0 - Successful                                                       #
#         1 - Failure                                                          #
################################################################################
# CALLS: S4P::logger                                                           #
#        S4P::perish                                                           #
#        wipe_request                                                          #
################################################################################
# CALLED BY: main                                                              #
################################################################################

    my ($dir, $datatype, $expiration_file) = @_;

    my ($file, $delete_failures);

    S4P::logger("DEBUG", "wipe_directory(): Entering wipe_directory() with dir: [$dir] and datatype: [$datatype]");

    $delete_failures = 0;

    opendir(DIR, $dir) or S4P::perish(70, "wipe_directory(): Cannot open directory [$dir]");

    while ( defined($file = readdir(DIR)) ) {
       next if ( $file eq "." or $file eq "..");
       if ( unlink("$dir/$file") ) {
          S4P::logger("DEBUG", "wipe_directory(): $file in directory $dir deleted.");
       } else {
          S4P::logger("WARNING", "wipe_directory(): Failed to delete $file.");
          $delete_failures++;
       }
    }

    if ( $delete_failures == 0 ) {
        if ( -e "$dir.met" ) {
            unlink("$dir.met") or 
                S4P::logger("WARNING", "wipe_directory(): Failed to delete associated ODL metadata file: $dir.met.");
        }
        if ( -e "$dir.xml" ) {
            unlink("$dir.xml") or 
                S4P::logger("WARNING", "wipe_directory(): Failed to delete associated XML metadata file: $dir.xml.");
        }
        if ( -e "$dir.ur" ) {
            wipe_request("$dir.ur", $datatype, $expiration_file);
            unlink("$dir.ur") or 
                S4P::logger("WARNING", "wipe_directory(): Failed to delete associated UR file: $dir.ur.");
        }
        if ( rmdir($dir) ) {
            S4P::logger("DEBUG", "wipe_directory(): directory: [$dir] deleted.");
        } else {
             S4P::logger("WARNING", "wipe_directory(): Failed to delete directory: $dir.");
             S4P::logger("DEBUG", "wipe_directory(): Leaving wipe_directory()");
             return 1;
        }
    } else {
         S4P::logger("WARNING", "wipe_directory(): Cannot delete directory $dir since one or files within cannot be deleted.");
         S4P::logger("DEBUG", "wipe_directory(): Leaving wipe_directory()");
         return 1;
    }

    S4P::logger("DEBUG", "wipe_directory(): Leaving wipe_directory()");
    return 0;

}

sub wipe_request {
 
################################################################################
#                                  wipe_request                                #
################################################################################
# PURPOSE: To delete a Request Data station request stub file                  #
################################################################################
# DESCRIPTION: wipe_request deletes the request stub file associated with the  #
#              file being deleted. Only files associated with data types being #
#              orderded via the Request Data station have request stub files.  #
################################################################################
# RETURN: 0 - Successful                                                       #
#         1 - Failure                                                          #
################################################################################
# CALLS: S4P::logger                                                           #
#        S4P::perish                                                           #
#        S4PM::parse_ur_file                                                   #
#        compare_urs                                                           #
################################################################################
# CALLED BY: wipe_file                                                         #
#            wipe_directory                                                    #
################################################################################

    my ($urfile, $datatype, $expiration_file) = @_;
    my $req;

    S4P::logger("DEBUG", "wipe_request(): UR file: $urfile");
    S4P::logger("DEBUG", "wipe_request(): data type: [$datatype]");

    open(UR, "$urfile") or S4P::perish(100, "wipe_request(): Cannot open UR file: $urfile: $!");
    
    my $file_ur = <UR>;
    chomp($file_ur);
    S4P::logger("DEBUG", "wipe_request(): UR is: [$file_ur]");
    close UR or S4P::perish(100, "wipe_request(): Cannot close UR file: $urfile: $!");

### If the UR is a "fake", LGID-style UR, then there should be no request
### stub file, so just return. -- C. Lynnes
    if ($file_ur =~ /^LGID/) {
        S4P::logger("INFO", "wipe_request(): LGID-style UR [$file_ur], so no corresponding request stub");
        return 0;
    }

### With the UR in hand, examine the request stub files and find the one
### whose contents matches the UR. This will be the stub file we want to
### delete.

    opendir(DIR, "$req_stub_dir") or S4P::perish(100, "wipe_requests(): Cannot open directory $req_stub_dir: $!");
 
    while ( defined($req = readdir(DIR)) ) {
        next if ( $req eq "." or $req eq "..");
        if ( $req =~ /^$datatype/ ) {
            my $status = open(REQ, "$req_stub_dir/$req");
            unless ( $status ) {
                S4P::logger("ERROR", "wipe_request(): Failed to open file $req_stub_dir/$req: $!");
                next;
            }
            my $req_ur = <REQ>;
            chomp($req_ur);
            close(REQ) or S4P::perish(100, "wipe_request(): Cannot close file: [$req_stub_dir/$req]: $!");
            if ( compare_urs($file_ur, $req_ur) ) {
                if ( unlink("$req_stub_dir/$req") ) {
                    S4P::logger("INFO", "wipe_request(): Request stub file: $req_stub_dir/$req deleted."); 
                    write_expiration($file_ur, $expiration_file) if $expiration_file;
                } else {
                    S4P::logger("ERROR", "wipe_request(): Failed to delete Request stub file: $req_stub_dir/$req: $!");
                }
                last;
            } else {
                S4P::logger("DEBUG", "wipe_request(): [$req_ur] is NOT equal to [$file_ur]. Therefore, NOT deleting a request stub file.");
            }
        }
    }

    closedir(DIR) or S4P::perish(70, "wipe_request(): Cannot closedir: [$req_stub_dir]: $!");
}
sub write_expiration {
    my ($ur, $expiration_file) = @_;
    open EXPIRE, ">>$expiration_file" or 
        S4P::logger('ERROR', "Cannot update expiration file $expiration_file: $!");
    print EXPIRE "$ur\n";
    close EXPIRE;
}
